@@ -19,7 +19,6 @@
 
 #include "../config.h"
 #include <sys/types.h>
-#include <stdio.h>
 #include <string.h>
 #include <glib-object.h>
 #include "debug.h"
@@ -29,32 +28,27 @@
 #include "trie.h"
 
 struct _vte_matcher {
-	gboolean initialized;
-	enum {
-		_vte_matcher_table,
-		_vte_matcher_trie
-	} type;
-	struct _vte_table *table;
-	struct _vte_trie *trie;
+	_vte_matcher_match_func match; /* shortcut to the most common op */
+	struct _vte_matcher_impl *impl;
+	GValueArray *free_params;
 };
 
 static GStaticMutex _vte_matcher_mutex = G_STATIC_MUTEX_INIT;
 static GCache *_vte_matcher_cache = NULL;
+static struct _vte_matcher_impl dummy_vte_matcher_trie = {
+	&_vte_matcher_trie
+};
+static struct _vte_matcher_impl dummy_vte_matcher_table = {
+	&_vte_matcher_table
+};
 
 /* Add a string to the matcher. */
 static void
-_vte_matcher_add(struct _vte_matcher *matcher,
+_vte_matcher_add(const struct _vte_matcher *matcher,
 		 const char *pattern, gssize length,
 		 const char *result, GQuark quark)
 {
-	switch (matcher->type) {
-	case _vte_matcher_table:
-		_vte_table_add(matcher->table, pattern, length, result, quark);
-		break;
-	case _vte_matcher_trie:
-		_vte_trie_add(matcher->trie, pattern, length, result, quark);
-		break;
-	}
+	matcher->impl->klass->add(matcher->impl, pattern, length, result, quark);
 }
 
 /* Loads all sequences into matcher */
@@ -68,39 +62,38 @@ _vte_matcher_init(struct _vte_matcher *matcher, const char *emulation,
 	char *stripped;
 	int i;
 
-#ifdef VTE_DEBUG
-	if (_vte_debug_on(VTE_DEBUG_LIFECYCLE)) {
-		fprintf(stderr, "_vte_matcher_init()\n");
-	}
-#endif
+	_vte_debug_print(VTE_DEBUG_LIFECYCLE, "_vte_matcher_init()\n");
 
-	/* Load the known capability strings from the termcap structure into
-	 * the table for recognition. */
-	for (i = 0;
-	     _vte_terminal_capability_strings[i].capability[0];
-	     i++) {
-		if (_vte_terminal_capability_strings[i].key) {
-			continue;
-		}
-		code = _vte_terminal_capability_strings[i].capability;
-		stripped = _vte_termcap_find_string_length(termcap,
-                                                           emulation,
-                                                           code,
-                                                           &stripped_length);
-		if (stripped[0] != '\0') {
-			_vte_matcher_add(matcher, stripped, stripped_length,
-					 code, 0);
-			if (stripped[0] == '\r') {
-				found_cr = TRUE;
-			} else
-			if (stripped[0] == '\n') {
-				if ((strcmp(code, "sf") == 0) ||
-				    (strcmp(code, "do") == 0)) {
-					found_lf = TRUE;
-				}
+	if (termcap != NULL) {
+		/* Load the known capability strings from the termcap
+		 * structure into the table for recognition. */
+		for (i = 0;
+				_vte_terminal_capability_strings[i].capability[0];
+				i++) {
+			if (_vte_terminal_capability_strings[i].key) {
+				continue;
 			}
+			code = _vte_terminal_capability_strings[i].capability;
+			stripped = _vte_termcap_find_string_length(termcap,
+					emulation,
+					code,
+					&stripped_length);
+			if (stripped[0] != '\0') {
+				_vte_matcher_add(matcher,
+						stripped, stripped_length,
+						code, 0);
+				if (stripped[0] == '\r') {
+					found_cr = TRUE;
+				} else
+					if (stripped[0] == '\n') {
+						if (strcmp(code, "sf") == 0 ||
+								strcmp(code, "do") == 0) {
+							found_lf = TRUE;
+						}
+					}
+			}
+			g_free(stripped);
 		}
-		g_free(stripped);
 	}
 
 	/* Add emulator-specific sequences. */
@@ -124,13 +117,11 @@ _vte_matcher_init(struct _vte_matcher *matcher, const char *emulation,
 		_vte_matcher_add(matcher, "\n", 1, "sf", 0);
 	}
 
-#ifdef VTE_DEBUG
-	if (_vte_debug_on(VTE_DEBUG_TRIE)) {
-		fprintf(stderr, "Trie contents:\n");
+	_VTE_DEBUG_IF(VTE_DEBUG_TRIE) {
+		g_printerr("Trie contents:\n");
 		_vte_matcher_print(matcher);
-		fprintf(stderr, "\n");
+		g_printerr("\n");
 	}
-#endif
 }
 
 /* Allocates new matcher structure. */
@@ -140,31 +131,17 @@ _vte_matcher_create(gpointer key)
 	char *emulation = key;
 	struct _vte_matcher *ret = NULL;
 
-#ifdef VTE_DEBUG
-	if (_vte_debug_on(VTE_DEBUG_LIFECYCLE)) {
-		fprintf(stderr, "_vte_matcher_create()\n");
-	}
-#endif
+	_vte_debug_print(VTE_DEBUG_LIFECYCLE, "_vte_matcher_create()\n");
 	ret = g_slice_new(struct _vte_matcher);
-	ret->initialized = FALSE;
-	ret->type = _vte_matcher_trie;
-	ret->table = NULL;
-	ret->trie = NULL;
+	ret->impl = &dummy_vte_matcher_trie;
+	ret->match = NULL;
+	ret->free_params = NULL;
 
 	if (strcmp(emulation, "xterm") == 0) {
-		ret->type = _vte_matcher_table;
+		ret->impl = &dummy_vte_matcher_table;
 	} else
 	if (strcmp(emulation, "dtterm") == 0) {
-		ret->type = _vte_matcher_table;
-	}
-
-	switch (ret->type) {
-	case _vte_matcher_table:
-		ret->table = _vte_table_new();
-		break;
-	case _vte_matcher_trie:
-		ret->trie = _vte_trie_new();
-		break;
+		ret->impl = &dummy_vte_matcher_table;
 	}
 
 	return ret;
@@ -176,17 +153,12 @@ _vte_matcher_destroy(gpointer value)
 {
 	struct _vte_matcher *matcher = value;
 
-#ifdef VTE_DEBUG
-	if (_vte_debug_on(VTE_DEBUG_LIFECYCLE)) {
-		fprintf(stderr, "_vte_matcher_destroy()\n");
-	}
-#endif
-	if (matcher->table != NULL) {
-		_vte_table_free(matcher->table);
-	}
-	if (matcher->trie != NULL) {
-		_vte_trie_free(matcher->trie);
+	_vte_debug_print(VTE_DEBUG_LIFECYCLE, "_vte_matcher_destroy()\n");
+	if (matcher->free_params != NULL) {
+		g_value_array_free (matcher->free_params);
 	}
+	if (matcher->match != NULL) /* do not call destroy on dummy values */
+		matcher->impl->klass->destroy(matcher->impl);
 	g_slice_free(struct _vte_matcher, matcher);
 }
 
@@ -204,15 +176,16 @@ _vte_matcher_new(const char *emulation, struct _vte_termcap *termcap)
 	if (_vte_matcher_cache == NULL) {
 		_vte_matcher_cache = g_cache_new(_vte_matcher_create,
 				_vte_matcher_destroy,
-			       	(GCacheDupFunc) g_strdup, g_free,
+				(GCacheDupFunc) g_strdup, g_free,
 				g_str_hash, g_direct_hash, g_str_equal);
 	}
 
 	ret = g_cache_insert(_vte_matcher_cache, (gpointer) emulation);
 
-	if (!ret->initialized) {
+	if (ret->match == NULL) {
+		ret->impl = ret->impl->klass->create();
+		ret->match = ret->impl->klass->match;
 		_vte_matcher_init(ret, emulation, termcap);
-		ret->initialized = TRUE;
 	}
 
 	g_static_mutex_unlock(&_vte_matcher_mutex);
@@ -236,54 +209,40 @@ _vte_matcher_match(struct _vte_matcher *matcher,
 		   const char **res, const gunichar **consumed,
 		   GQuark *quark, GValueArray **array)
 {
-	switch (matcher->type) {
-	case _vte_matcher_table:
-		return _vte_table_match(matcher->table, pattern, length,
-					res, consumed, quark, array);
-		break;
-	case _vte_matcher_trie:
-		return _vte_trie_match(matcher->trie, pattern, length,
-				       res, consumed, quark, array);
-		break;
+	if (G_UNLIKELY (array != NULL && matcher->free_params != NULL)) {
+		*array = matcher->free_params;
+		matcher->free_params = NULL;
 	}
-	return NULL;
+	return matcher->match(matcher->impl, pattern, length,
+					res, consumed, quark, array);
 }
 
 /* Dump out the contents of a matcher, mainly for debugging. */
 void
 _vte_matcher_print(struct _vte_matcher *matcher)
 {
-	switch (matcher->type) {
-	case _vte_matcher_table:
-		_vte_table_print(matcher->table);
-		break;
-	case _vte_matcher_trie:
-		_vte_trie_print(matcher->trie);
-		break;
-	}
+	matcher->impl->klass->print(matcher->impl);
 }
 
 /* Free a parameter array.  Most of the GValue elements can clean up after
  * themselves, but we're using gpointers to hold unicode character strings, and
  * we need to free those ourselves. */
 void
-_vte_matcher_free_params_array(GValueArray *params)
+_vte_matcher_free_params_array(struct _vte_matcher *matcher,
+		               GValueArray *params)
 {
 	guint i;
-	GValue *value;
-	gpointer ptr;
-	if (params != NULL) {
-		for (i = 0; i < params->n_values; i++) {
-			value = g_value_array_get_nth(params, i);
-			if (G_VALUE_HOLDS_POINTER(value)) {
-				ptr = g_value_get_pointer(value);
-				if (ptr != NULL) {
-					g_free(ptr);
-				}
-				g_value_set_pointer(value, NULL);
-			}
+	for (i = 0; i < params->n_values; i++) {
+		GValue *value = &params->values[i];
+		if (G_UNLIKELY (g_type_is_a (value->g_type, G_TYPE_POINTER))) {
+			g_free (g_value_get_pointer (value));
 		}
-		g_value_array_free(params);
+	}
+	if (G_UNLIKELY (matcher == NULL || matcher->free_params != NULL)) {
+		g_value_array_free (params);
+	} else {
+		matcher->free_params = params;
+		params->n_values = 0;
 	}
 }
 
